#!/usr/bin/env python
import contextlib
import josepy
import logging
import os
import psycopg2
import time

from acme import challenges
from acme import client as acme_client
from acme import crypto_util
from acme import messages
from acme.errors import ValidationError

from cryptography import x509
from cryptography.x509.oid import NameOID, ExtensionOID
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa

# Account key size
ACC_KEY_BITS = 2048

# Certificate private key size
CERT_PKEY_BITS = 2048

LE_DIRECTORIES = {
    "staging-v02": "https://acme-staging-v02.api.letsencrypt.org/directory",
    "production-v02": "https://acme-v02.api.letsencrypt.org/directory",
}
ACME_API_VERSION = "v02"

_logger = logging.getLogger(__name__)


def _directory(env):
    try:
        directory_url = LE_DIRECTORIES["{}-{}".format(env, ACME_API_VERSION)]
    except KeyError:
        raise Exception("invalid env %s" % env)
    return directory_url


def _account_config(env):
    conf_dir = os.path.join(os.path.expanduser('~'), 'config', 'certtool', env)
    pkey_path = os.path.join(conf_dir, 'pkey.pem')
    regr_path = os.path.join(conf_dir, 'regr.json')
    return conf_dir, pkey_path, regr_path


def _init_account(env, email=None):
    """Generate a new random key, use it to build an acme.Client and register a
    new account. Finally store the key and registration info for later use."""
    pkey, pkey_pem = _generate_rsa_key(ACC_KEY_BITS)

    pkey_jwk = josepy.JWKRSA(key=pkey)
    net = acme_client.ClientNetwork(pkey_jwk)
    directory_url = _directory(env)
    directory = messages.Directory.from_json(net.get(directory_url).json())
    client = acme_client.ClientV2(directory, net=net)

    _logger.info("certtool: registering new account, env: %s", env)
    regr = client.new_account(messages.NewRegistration.from_data(
        email=email, terms_of_service_agreed=True))

    conf_dir, pkey_path, regr_path = _account_config(env)
    if not os.path.isdir(conf_dir):
        os.makedirs(conf_dir)

    pkey_f = open(pkey_path, 'wb+')
    regr_f = open(regr_path, 'w+')

    pkey_f.write(pkey_pem)
    regr_f.write(regr.json_dumps())

    regr_f.close()
    pkey_f.close()
    return client


def _generate_rsa_key(bit_length):
    pkey = rsa.generate_private_key(65537, bit_length, default_backend())
    pkey_pem = pkey.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption())
    return pkey, pkey_pem


def new_pkey_csr(domains, pkey_pem=None):
    """Create certificate signing request."""
    if pkey_pem is None:
        # Create private key.
        _, pkey_pem = _generate_rsa_key(CERT_PKEY_BITS)
    csr_pem = crypto_util.make_csr(pkey_pem, domains)
    return pkey_pem, csr_pem


def make_client(env, email=None):
    """ Build an acme.Client using locally stored account config. Register a
    new account if no such config is found. """
    conf_dir, pkey_path, regr_path = _account_config(env)
    if not os.path.isdir(conf_dir):
        return _init_account(env, email=email)

    pkey_f = open(pkey_path, 'rb')
    regr_f = open(regr_path, 'r')

    pkey = serialization.load_pem_private_key(pkey_f.read(), password=None, backend=default_backend())
    key_jwk = josepy.JWKRSA(key=pkey)
    regr = messages.RegistrationResource.json_loads(regr_f.read())

    regr_f.close()
    pkey_f.close()

    _logger.info("certtool: building acme client, env: %s", env)
    net = acme_client.ClientNetwork(key_jwk)
    net.account = regr
    directory_url = _directory(env)
    directory = messages.Directory.from_json(net.get(directory_url).json())
    client = acme_client.ClientV2(directory, net=net)
    return client


def _remove_dns01_response(cr, domain):
    query = """
        DELETE FROM records
         WHERE name = %s
           AND type = 'TXT'
           AND origin = 'acme'"""
    params = (domain,)
    cr.execute(query, params)


def _add_dns01_response(cr, name, value):
    query = """
        INSERT INTO records (name, content, type, ttl, prio, origin, domain_id)
        VALUES (%s, %s, 'TXT', 1, 0, 'acme', 1)"""
    params = (name, value)
    cr.execute(query, params)


def _remove_http01_response(cr, domain):
    query = """
        UPDATE ssl_certificate
           SET http_01_validation_path = NULL, http_01_validation = NULL
         WHERE subject_cn = %s
    """
    params = (domain,)
    cr.execute(query, params)


def _add_http01_response(cr, domain, validation_path, validation_token):
    query = """
        UPDATE ssl_certificate
           SET http_01_validation_path = %s, http_01_validation = %s
         WHERE subject_cn = %s
    """
    params = (validation_path, validation_token, domain)
    cr.execute(query, params)


def get_chall(authz, typ):
    for chall_body in authz.body.challenges:
        if isinstance(chall_body.chall, typ):
            return chall_body
    raise Exception("No %s challenge found" % typ)


def do_dns_challenges(client, authzs, cr=None):
    cleanup_hosts = []
    for a in authzs:
        c = get_chall(a, challenges.DNS01)
        name, value = (
            c.validation_domain_name(a.body.identifier.value),
            c.validation(client.net.key))
        cleanup_hosts.append(name)
        with records_cursor(cr) as cursor:
            _add_dns01_response(cursor, name, value)
        _logger.info("certtool: dns-01 challenge: added value %s for %s",
                     value, name)
        # LE recursors do not necessarily see DNS updates as soon as they are
        # visible at our nameservers (despite the synchronous replication with
        # remote_apply level).
        time.sleep(3)
        client.answer_challenge(c, c.response(client.net.key))

    def cleanup():
        with records_cursor(cr) as cursor:
            for host in cleanup_hosts:
                _remove_dns01_response(cursor, host)
        _logger.info("certtool: dns-01 challenge: cleaned up hosts %s",
                     cleanup_hosts)
    return cleanup


def do_http_challenge(client, authz, domain, cr=None):
    c = get_chall(authz, challenges.HTTP01)
    validation_path, validation_token = (
        c.path.split('/')[-1],
        c.validation(client.net.key))

    with records_cursor(cr) as cursor:
        _add_http01_response(cursor, domain, validation_path, validation_token)
    client.answer_challenge(c, c.response(client.net.key))

    def cleanup():
        with records_cursor(cr) as cursor:
            _remove_http01_response(cursor, domain)
    return cleanup


def x509_info(x509_pem):
    cert = x509.load_pem_x509_certificate(x509_pem, default_backend())
    subject_cn = cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value
    ext = cert.extensions.get_extension_for_oid(ExtensionOID.SUBJECT_ALTERNATIVE_NAME)
    subject_alt_name = ext.value.get_values_for_type(x509.DNSName)
    return {
        'subject_cn': subject_cn,
        'subject_alt_name': subject_alt_name,
        'not_valid_before': cert.not_valid_before,
        'not_valid_after': cert.not_valid_after,
    }


def auth_and_issue(domains, chall_type="dns-01", email=None, env="staging", cr=None):
    """Make authzs for each of the given domains, tell the ACME server to
    validate the challenges, then poll for the authzs to be ready and issue a
    cert."""
    client = make_client(env, email)

    pkey_pem, csr_pem = new_pkey_csr(domains)
    orderr = client.new_order(csr_pem)
    authzs = orderr.authorizations

    _logger.info("certtool: starting %s challenge for domains: %s", chall_type, domains)
    if chall_type == "http-01":
        cleanup = do_http_challenge(client, authzs[0], domains[0], cr=cr)
    elif chall_type == "dns-01":
        cleanup = do_dns_challenges(client, authzs, cr=cr)
    else:
        raise Exception("invalid challenge type %s" % chall_type)

    try:
        finalized_orderr = client.poll_and_finalize(orderr)
        fullchain_pem = finalized_orderr.fullchain_pem.encode('ascii')
        _logger.info("certtool: successful authorization for domains: %s", domains)
        return pkey_pem, fullchain_pem, x509_info(fullchain_pem)
    except ValidationError as e:
        err_details = []
        for authr in e.failed_authzrs:
            for c in authr.body.challenges:
                if c.error:
                    err_details += [c.error.detail]
        raise Exception("ACME error: authorization failure(s): " + "; ".join(err_details))
    finally:
        cleanup()


@contextlib.contextmanager
def records_cursor(cr=None):
    """Yield [Odoo] cursor if given, yield raw cursor otherwise. Allows us to
    work outside of an Odoo context."""
    if cr:
        yield cr
        cr.commit()
    else:
        conn = psycopg2.connect("dbname=saas_master")
        yield conn.cursor()
        conn.commit()


if __name__ == '__main__':
    """
    Standalone behavior (not completely implemented)
    """
    domain_list = [['*.test.odoo.com', 'test.odoo.com']]
    email = "certs-saas@odoo.com"
    cwd = os.path.dirname(os.path.realpath(__file__))
    for domains in domain_list:
        pkey_pem, fullchain_pem, _ = auth_and_issue(domains, email=email)
        base_fname = domains[0].replace('*', '').replace('.', '_')
        crt_path = os.path.join(cwd, base_fname + '.crt')
        key_path = os.path.join(cwd, base_fname + '.key')
        cert_fd = open(crt_path, 'wb+')
        key_fd = open(key_path, 'wb+')
        try:
            key_fd.write(pkey_pem)
            cert_fd.write(fullchain_pem)
        finally:
            cert_fd.close()
            key_fd.close()
